import pydantic

from captur_ml_sdk.dtypes import ImageClassificationResult
from captur_ml_sdk.dtypes.generics import Endpoint
from typing import List


class LivePredictionRequest(pydantic.BaseModel):
    endpoint: Endpoint = pydantic.Field(
        ...,
        description="The ID and location of the VertexAI endpoint to use for prediction.",
    )
    image_url: pydantic.AnyUrl = pydantic.Field(
        ...,
        description="The URL of the image to use for prediction.",
    )
    request_id: str = pydantic.Field(
        "",
        description="Unique ID for the request."
    )
    webhooks: List[pydantic.AnyUrl] = pydantic.Field(
        ...,
        description="List of webhooks to which the prediction results will be sent.",
    )

class LivePredictionResponse(pydantic.BaseModel):
    endpoint: Endpoint = pydantic.Field(
        ...,
        description="The ID and location of the VertexAI endpoint to use for prediction.",
    )
    images: List[ImageClassificationResult] = pydantic.Field(
        ...,
        description='The list of classification results of the images used for the prediction.'
    )
    request_id: str = pydantic.Field(
        ...,
        description="Unique ID for the request."
    )



from typing import Union, List

from captur_ml_sdk.dtypes import ImageClassificationResult
from captur_ml_sdk.dtypes.generics import Image
from captur_ml_sdk._services.aiplatform.live_predict import get_image_classification_prediction_from_deployed_automl


def get_live_automl_image_classification(
    endpoint_id: str,
    location: str,
    api_endpoint: str,
    images: Union[Image, List[Image]],
    return_raw_automl_response: bool = False
):
    """Perform image classification on an image or list of images using AutoML.

    Args:
        endpoint_id (str): The VertexAI endpoint of the AutoML model that will perform the classification.
        location (str) : The location of the VertexAI endpoint.
        api_endpoint (str): The endpoint provided by the client.
        images (Union[Image, List[Image]]): The image(s) to classify.
        return_raw_automl_response (bool, optional): If True, AutoML response will not be cast to
            captur_ml_sdk.dtypes.ImageClassificationResult. Defaults to False.

    Returns:
        ImageClassificationResult: Result of the model prediction
    """
    if isinstance(images, Image):
        images = [images]

    image_classification_raw_results = [
        (
            image, get_image_classification_prediction_from_deployed_automl(
                endpoint_id=endpoint_id,
                location=location,
                api_endpoint=api_endpoint,
                image_data=image.data,
            )
        ) for image in images
    ]

    if return_raw_automl_response:
        return image_classification_raw_results

    image_classification_results = [
        ImageClassificationResult.from_automl_response(
            image, image_classification_raw_result
        ) for image, image_classification_raw_result in image_classification_raw_results
    ]

    return image_classification_results



import pydantic

from typing import Optional

class Endpoint(pydantic.BaseModel):
    id: str = pydantic.Field(
        ...,
        description='A unique identifier for the VertexAI Endpoint.'
    )
    location: str = pydantic.Field(
        ...,
        description='The global location of the VertexAI Endpoint.'
    )



import re
from captur_ml_sdk.dtypes.exceptions import (
    GoogleCloudVertexAIEndpointDoesNotExistError,
    GoogleCloudVertexAIEndpointImageTooLargeError,
    GoogleCloudVertexAIEndpointNoModelDeployedError,
)

from google.cloud import aiplatform_v1
from google.cloud.aiplatform.gapic import PredictionServiceClient, schema
from google.api_core import exceptions as google_exceptions


def get_image_classification_prediction_from_deployed_automl(
    image_data: str,
    endpoint_id: str,
    project: str = "capturpwa",
    location: str = "us-central1",
    api_endpoint: str = "us-central1-aiplatform.googleapis.com",
) -> aiplatform_v1.types.PredictResponse:
    """Gets image classification prediction from deployed automl model.

    Args:
        image_data (str): utf-8 encoded string of image.
        endpoint_id (str): The id that uniquely identifies the GCP Vertex AI endpoint at which the model is deployed.
        project (str, optional): The GCP project name. Defaults to "capturpwa".
        location (str, optional): The global location where the prediction computation takes place. Defaults to "us-central1".
        api_endpoint (str, optional): The endpoint provided by the client. Defaults to "us-central1-aiplatform.googleapis.com".

    Raises:
        GoogleCloudVertexAIEndpointDoesNotExistError: The specified endpoint does not exist.
        GoogleCloudVertexAIEndpointNoModelDeployedError: There is no model deployed at the specified endpoint.
        GoogleCloudVertexAIEndpointImageTooLargeError: The image exceeds the 1.5MB limit.

    Returns:
        automl_v1beta1.types.PredictResponse: An object containing the prediction results for the image.
    """
    client_options = {"api_endpoint": api_endpoint}
    client = PredictionServiceClient(client_options=client_options)

    instance = schema.predict.instance.ImageClassificationPredictionInstance(
        content=image_data,
    ).to_value()
    instances = [instance]

    parameters = schema.predict.params.ImageClassificationPredictionParams(
        confidence_threshold=0.0, max_predictions=10,
    ).to_value()

    endpoint = client.endpoint_path(
        project=project, location=location, endpoint=endpoint_id
    )
    try:
        response = client.predict(
            endpoint=endpoint, instances=instances, parameters=parameters
        )
    except google_exceptions.NotFound:
        raise GoogleCloudVertexAIEndpointDoesNotExistError(f"Endpoint {endpoint_id} not found.")

    except google_exceptions.FailedPrecondition as e:
        if "doesn't have traffic_split" in str(e):
            raise GoogleCloudVertexAIEndpointNoModelDeployedError(f"No model is deployed at endpoint {endpoint_id}.")
        elif re.search(r"exceeds \d*.* limit", str(e)) is not None:
            match = re.search(r"exceeds (\d*.*) limit", str(e))
            size_limit = match.group(1)
            raise GoogleCloudVertexAIEndpointImageTooLargeError(size_limit)

    except google_exceptions.InvalidArgument as e:
        if "400 Provided image is not valid" in str(e):
            raise ValueError(str(e))
    return response



# ============ PubSub Exceptions ============ #
class GoogleCloudPubSubTopicDoesNotExistError(Exception):
    pass

# ============ Google Cloud Storage Exceptions ============ #
class GoogleCloudStoragePermissionError(Exception):
    pass
class GoogleCloudStorageResourceNotFoundError(Exception):
    pass
class GoogleCloudStorageBucketNotFoundError(Exception):
    pass

# ============ Google Cloud Vertex AI Endpoint Exceptions ============ #
class GoogleCloudVertexAIEndpointDoesNotExistError(Exception):
    pass
class GoogleCloudVertexAIEndpointImageTooLargeError(Exception):
    pass
class GoogleCloudVertexAIEndpointNoModelDeployedError(Exception):
    pass

# ============ Google Cloud Vertex AI General Exceptions ============ #
class GoogleCloudVertexAIModelDoesNotExistError(Exception):
    pass
class GoogleCloudVertexAIDatasetDoesNotExistError(Exception):
    pass
class GoogleCloudVertexAIResourceDoesNotExistError(Exception):
    pass
class GoogleCloudVertexAICompatibilityError(Exception):
    pass

# ============ Sentry Errors ============ #
class SentryDSNNotProvidedError(Exception):
    pass



import base64
import json
import pydantic
import requests

from typing import Dict, Union


def send_payload_to_webhook(webhook_url: str, payload: Union[Dict, pydantic.BaseModel]):
    """Sends a payload to a webhook.

    Args:
        webhook_url (str): URL of the webhook to send the event to.
        payload [Dict, pydantic.BaseModel]: Content to send to the webhook.

    Raises:
        ValueError: Raised if the payload is not a dictionary or pydantic.BaseModel
                    the post request fails.
    """
    if not isinstance(payload, dict):
        if isinstance(payload, pydantic.BaseModel):
            payload = payload.dict()
        else:
            raise ValueError("Payload must be a dictionary or a pydantic model.")
    try:
        requests.post(webhook_url, json=payload)
    except Exception:
        raise ValueError(f"Unable to post to {webhook_url}")


def push_published_message_to_webhooks(event, context):
    """Background Cloud Function to be triggered by Pub/Sub.
    Args:
         event (dict):  The dictionary with data specific to this type of
                        event. The `@type` field maps to
                         `type.googleapis.com/google.pubsub.v1.PubsubMessage`.
                        The `data` field maps to the PubsubMessage data
                        in a base64-encoded string. The `attributes` field maps
                        to the PubsubMessage attributes if any is present.
         context (google.cloud.functions.Context): Metadata of triggering event
                        including `event_id` which maps to the PubsubMessage
                        messageId, `timestamp` which maps to the PubsubMessage
                        publishTime, `event_type` which maps to
                        `google.pubsub.topic.publish`, and `resource` which is
                        a dictionary that describes the service API endpoint
                        pubsub.googleapis.com, the triggering topic's name, and
                        the triggering event type
                        `type.googleapis.com/google.pubsub.v1.PubsubMessage`.
    Returns:
        None. The output is written to Cloud Logging.
    """
    pass
    event_data = json.loads(base64.b64decode(event['data']).decode('utf-8'))
    webhooks = event_data.get("webhooks")
    if not isinstance(webhooks, list):
        webhooks = [webhooks]
    event_content = event_data["content"]

    if not event_data:
        event_data = {
            "example_data": "Here is some example data that will in the future be read "
            "from the event message."
        }

    for url in webhooks:
        # This is where we actually make the requests aka trigger the relevant webhooks
        send_payload_to_webhook(url, event_content)

    return {
        "urls": webhooks,
        "body": event_content
    }



from google.cloud import aiplatform
from google.api_core import exceptions as google_exceptions
from captur_ml_sdk.dtypes.exceptions import (
    GoogleCloudVertexAIModelDoesNotExistError,
    GoogleCloudVertexAIResourceDoesNotExistError,
    GoogleCloudStorageBucketNotFoundError,
)
from captur_ml_sdk._services.storage import check_file_exists
from captur_ml_sdk.utils.file_formatting import (
    get_bucket_from_gs_uri, get_filepath_from_gs_uri
)

def make_batch_prediction(
    model_id: str,
    job_display_name: str,
    gcs_source: str,
    gcs_destination: str,
    project: str = "capturpwa",
    location: str = "us-central1",
    sync: bool = True,
):
    """Sends request to perform a batch prediction job.

    Args:
        model_id (str): GCP model id that uniquely identifies a machine learning model.
        job_display_name (str): the identifier that will appear in the batch prediction dashboard on GCP
        gcs_source (str): GCS filepath of the batch file. The batch file is a JSONL file that lists the image
            Google Storage URIs to be predicted.
        gcs_destination (str): GCS filepath to store outputted prediction JSON files.
        project (str, optional): The GCP project name. Defaults to "capturpwa".
        location (str, optional): The geographical location where the prediction service is performed. Should be
            the same location as the model specified by `model_id`. Defaults to "us-central1".
        sync (bool, optional): whether the batch prediction is performed synchronously or asynchronously. Defaults to True.

    Raises:
        GoogleCloudVertexAIModelDoesNotExistError: No model is found with id "model_id".
        GoogleCloudVertexAIBatchGCSSourceDoesNotExistError: Source batch file is not found.
        GoogleCloudStorageBucketNotFoundError: A Google Cloud Storage bucket is not found.
    """

    aiplatform.init(project=project, location=location)

    try:
        my_model = aiplatform.Model(model_id)
    except google_exceptions.NotFound:
        raise GoogleCloudVertexAIModelDoesNotExistError(f"Model {model_id} not found.")

    for error_detail, gcs_file in [("source", gcs_source), ("destination", gcs_destination)]:
        try:
            bucket = get_bucket_from_gs_uri(gcs_file)
            filepath = get_filepath_from_gs_uri(gcs_file)
            if not check_file_exists(bucket, filepath):
                raise GoogleCloudVertexAIResourceDoesNotExistError(f"GCS {error_detail} file '{gcs_file}' not found.")
        except GoogleCloudStorageBucketNotFoundError:
            raise GoogleCloudStorageBucketNotFoundError(f"GCS {error_detail} bucket '{gcs_file}' not found.")

    if len(job_display_name) > 128:
        raise ValueError("job_display_name must be 128 characters or less.")

    return my_model.batch_predict(
        job_display_name=job_display_name,
        gcs_source=gcs_source,
        gcs_destination_prefix=gcs_destination,
        sync=sync,
    )



class GsUri(object):
    def __init__(self, gs_uri: str):
        """Object representing a Google Storage URI.

        Args:
            gs_uri (str): Google Storage URI.

        Raises:
            ValueError: Raised if gs_uri is not a valid Google Storage URI.
        """
        if not gs_uri.startswith("gs://"):
            raise ValueError(f"{gs_uri} is not a valid Google Storage URI.")
        self.gs_uri = gs_uri

    def get_bucket(self) -> str:
        """Extracts the Bucket name.

        Returns:
            str: Google Storage Bucket.
        """
        return self.gs_uri.split("/")[2]

    def get_filepath(self) -> str:
        """Extracts the filepath after the Bucket name.

        Returns:
            str: Filepath of the Google Storage URI.
        """
        return "/".join(self.gs_uri.split("/")[3:])


class GsHttp(object):
    def __init__(self, http_url: str):
        """Object representing a HTTP(S) URL to a file in Google Storage.

        Args:
            http_url (str): HTTP URL to file in Google Storage.

        Raises:
            ValueError: Raised if http_url is not a valid HTTP URL.
        """
        if not http_url.startswith("http://storage.googleapis.com/")\
                and not http_url.startswith("https://storage.googleapis.com/"):
            raise ValueError(f"{http_url} is not a valid HTTP URL.")
        self.http_url = http_url

    def convert_to_gs_uri(self) -> str:
        """Converts the HTTP URL to a Google Storage URI.

        Returns:
            str: Google Storage URI.
        """
        gs_uri = self.http_url.replace("https://storage.googleapis.com/", "gs://")
        return gs_uri

    def get_bucket(self) -> str:
        """Extracts the Bucket name.

        Returns:
            str: Google Storage Bucket.
        """
        return self.http_url.split("/")[2]

    def get_filepath(self) -> str:
        """Extracts the filepath after the Bucket name.

        Returns:
            str: Filepath of the Google Storage URI.
        """
        return "/".join(self.gs_uri.split("/")[3:])



from google.cloud import aiplatform
from typing import Optional

from captur_ml_sdk.dtypes.exceptions import (
    GoogleCloudVertexAIDatasetDoesNotExistError,
    GoogleCloudVertexAIModelDoesNotExistError,
    GoogleCloudVertexAICompatibilityError
)
from google.api_core import exceptions as google_exceptions


def train_model(
    location: str,
    display_name: str,
    dataset_id: str,
    prediction_type: str,
    project: str = "capturpwa",
    model_type: str = "CLOUD",
    model_display_name: Optional[str] = None,
    multi_label: bool = False,
    base_model: Optional[str] = None,
    training_fraction_split: float = 0.8,
    validation_fraction_split: float = 0.1,
    test_fraction_split: float = 0.1,
    budget_milli_node_hours: int = 8000,
    disable_early_stopping: bool = False,
    sync: bool = True,
    create_request_timeout: float = 300
):

    aiplatform.init(project=project, location=location)

    try:
        if base_model:
            base_model = aiplatform.Model(base_model)
            if prediction_type == "classification" and base_model.predict_schemata.prediction_schema_uri != "'https://storage.googleapis.com/google-cloud-aiplatform/schema/predict/prediction/classification_1.0.0.yaml'":
                raise GoogleCloudVertexAICompatibilityError("Base model must be type classification. Object detecttion does not support the use of a base model.")
    except google_exceptions.NotFound:
        raise GoogleCloudVertexAIModelDoesNotExistError(f"Base model {base_model} not found.")

    if len(model_display_name) > 128:
        raise ValueError("job_display_name must be 128 characters or less.")

    if create_request_timeout > 300:
        raise ValueError("Request timeout must be 300 seconds or less.")

    job = aiplatform.AutoMLImageTrainingJob(
        display_name=display_name,
        base_model=base_model,
        prediction_type=prediction_type,
        multi_label=multi_label,
        model_type=model_type,
    )

    try:
        my_image_ds = aiplatform.ImageDataset(dataset_id)
    except google_exceptions.NotFound:
        raise GoogleCloudVertexAIDatasetDoesNotExistError(
            f"Dataset {dataset_id} does not exist"
        )

    try:
        model = job.run(
            dataset=my_image_ds,
            model_display_name=model_display_name,
            training_fraction_split=training_fraction_split,
            validation_fraction_split=validation_fraction_split,
            test_fraction_split=test_fraction_split,
            budget_milli_node_hours=budget_milli_node_hours,
            disable_early_stopping=disable_early_stopping,
            sync=sync,
            create_request_timeout=create_request_timeout
        )
    except (
        google_exceptions.InvalidArgument,
        google_exceptions.FailedPrecondition
    ) as e:
        if "The fraction values must sum to up to 1" in str(e):
            raise ValueError(
                f"Training fraction: {training_fraction_split}, validation fraction: {validation_fraction_split}, and test fraction: {test_fraction_split} do not sum to 1."
            )
        elif "The train budget milli node hours must be in the range [20000, 900000]" in str(e):
            model_type = str(e).split(" ")[-1]
            raise ValueError(
                f"Training budget milli node hours must be in the range [20000, 900000] for prediction type {prediction_type} and model type {model_type}."
            )



import io
import base64
import urllib

from PIL import Image as PILImage
from typing import Union
from urllib.error import URLError

from captur_ml_sdk.dtypes.generics import Image

def resize_image_by_factor(image: PILImage.Image, factor: float):
    """Changes the size of the image by a factor, preserving the aspect ratio.

    Args:
        image (PIL Image): Image to be resized.
        factor (float): Factor that both width and height are multiplied by.

    Returns:
        image (PIL Image): resized image.
    """
    if not isinstance(image, PILImage.Image):
        input_type = type(image)
        raise TypeError(f"image is of type {input_type} but must be a PIL Image")
    if not isinstance(factor, float):
        input_type = type(factor)
        raise TypeError(f"factor is of type {input_type} but must be a float")
    w, h = image.size
    neww, newh = int(w * factor), int(h * factor)
    image = image.resize((neww, newh))
    return image

def convert_pil_image_to_bytes_array(image: PILImage.Image, encode_as_utf8: bool = False, image_format: str = "jpeg"):
    """Converts PIL image to bytes array.

    Args:
        image (PIL.Image): Image to be converted to a bytes array.

    Returns:
        bytes: Bytes array of the image.
    """
    if not isinstance(image, PILImage.Image):
        input_type = type(image)
        raise TypeError(f"image is of type {input_type} but must be a PIL Image")
    imgByteArr = io.BytesIO()
    image.save(imgByteArr, format=image_format)
    image_data = imgByteArr.getvalue()

    if encode_as_utf8:
        image_data = base64.b64encode(image_data).decode("utf-8")

    return image_data

def load_image_from_url(image_url: str, limit_image_size_to: int = 0, return_pil: bool = False) -> Union[PILImage.Image, Image]:
    """Retrieve an image via HTTP/S.

    Args:
        url (string): The http[s]://___ URL hosting the pure image data (i.e. not an HTML webpage, for example)
        limit_image_size_to (int, optional): If bigger than 0, limit the number of bytes of the image by rescaling. For example,
            a value of 1024 will reduce the image size (and quality) to fit into 1kB. NOTE: If this value is not 0, the image
            may be converted to JPEG depending on whether or not the resize operation is performed; if this is undesirable, set
            `return_pil` to True and convert to bytestring manually. Defaults to 0.
        return_pil (bool, optional): If True a PIL.Image instance will be returned, if False a captur_ml_sdk.dtypes.Image containing
            the UTF8-encoded string of the image data will be returned. Defaults to False.

    Raises:
        URLError: If the image cannot be retrieved from the URL.

    Returns:
        Image (Union[PIL.Image.Image, captur_ml_sdk.dtypes.Image]): If `return_pil` is True, a PIL.Image instance of the image is 
            returned, otherwise a captur_ml_sdk.dtypes.Image is returned.
    """
    try:
        with urllib.request.urlopen(image_url) as u:
            content_length = u.info().get(name="Content-Length")

            if content_length is None:
                raise URLError()

            image_size = int(content_length)
            rescale_factor = 0 if limit_image_size_to <= 0 or image_size < limit_image_size_to else limit_image_size_to / image_size

            if rescale_factor == 0:
                if return_pil:
                    return PILImage.open(u)

                image_data = u.read()
                image_data = base64.b64encode(image_data).decode("utf-8")
                image_data = Image(url=image_url, data=image_data)
            else:
                image_data = PILImage.open(u)
                image_data = resize_image_by_factor(image_data, rescale_factor)

                if return_pil:
                    return image_data

                image_data = convert_pil_image_to_bytes_array(image_data, encode_as_utf8=True)
                image_data = Image(url=image_url, data=image_data)

            return image_data
    except URLError:
        raise URLError(f"Could not retrieve image from url: {image_url}")



import pydantic

from typing import Optional

class Image(pydantic.BaseModel):
    id: str = pydantic.Field(
        "",
        description='A unique identifier of the image'
    )
    url: Optional[pydantic.AnyUrl] = pydantic.Field(
        "",
        description='An http[s]:// or gs:// URL pointing to the image resource. Required if `data` is not specified.'
    )
    data: Optional[str] = pydantic.Field(
        "",
        description='The image bytes encoded as a UTF-8 string. Required if `url` is not specified.'
    )

    @pydantic.root_validator
    def must_contain_url_or_data(cls, values):
        if not values.get("url") and not values.get("data"):
            raise ValueError("Either `url` or `data` must be specified.")
        return values



import base64
import json
import pydantic


def parse_data_from_pubsub_message(
    event: bytes,
    pydantic_class: pydantic.main.ModelMetaclass
):
    """Parses data from a Pub/Sub message and raises an error if the
       message has an invalid format.

    Args:
        event (bytes): Base64-encoded Pub/Sub message.
        pydantic_class (pydantic.main.ModelMetaclass): A specified pydantic class.

    Raises:
        ValueError: Raised if the message has an invalid format.

    Returns:
        PydanticClass: An instance of the specified pydantic class.
    """
    data = json.loads(base64.b64decode(event['data']).decode('utf-8'))

    err = None
    try:
        event_data = pydantic.parse_obj_as(pydantic_class, data)
    except pydantic.ValidationError as e:
        err = e.errors()

    if err is not None:
        raise ValueError(f"Errors: {err}")

    return event_data



import os
import sentry_sdk

from google.cloud import error_reporting
from captur_ml_sdk.dtypes.exceptions import SentryDSNNotProvidedError
from sentry_sdk.integrations.gcp import GcpIntegration


def get_google_cloud_error_reporting_client() -> error_reporting.Client:
    """Returns a Google Cloud error reporting client

    Returns:
        error_reporting.Client: A Google Cloud error reporting client
    """
    return error_reporting.Client()


def initialise_sentry(environment="development", gcp_integration=False):
    """Initialises the Sentry client according to the SENTRY_DSN environment variable

    Raises:
        SentryDSNNotProvidedError: If the SENTRY_DSN environment variable is not set
    """
    if not os.getenv("SENTRY_DSN"):
        raise SentryDSNNotProvidedError("SENTRY_DSN environment variable must be set to use Sentry for error reporting.")

    if gcp_integration:
        sentry_sdk.init(
            dsn=os.getenv("SENTRY_DSN"),
            environment=environment,
            integrations=[GcpIntegration(timeout_warning=True)],
            traces_sample_rate=1.0
        )
    else:
        sentry_sdk.init(
            dsn=os.getenv("SENTRY_DSN"),
            environment=environment,
            traces_sample_rate=1.0
        )



import pydantic

class Prediction(pydantic.BaseModel):
    id: str = pydantic.Field(
        "",
        description='A unique identifier of the Prediction'
    )
    displayName: str = pydantic.Field(
        ...,
        description='A human-readable name of the Prediction'
    )
    confidence: float = pydantic.Field(
        ...,
        description='The statistical confidence of the Prediction'
    )



from google.cloud import aiplatform
from typing import Sequence


def get_image_dataset(
    dataset_id: str,
    project: str = "capturpwa",
    location: str = "us-central1"
) -> aiplatform.ImageDataset:
    aiplatform.init(project=project, location=location)
    ds = aiplatform.ImageDataset(dataset_id)
    return ds


def export_data_files(dataset: aiplatform.ImageDataset, output_dir: str) -> Sequence[str]:
    data_files = dataset.export_data(output_dir=output_dir)
    return data_files


class AIPlatformImageDataset():
    def __init__(self, dataset_id: str, project: str = "capturpwa", location: str = "us-central1"):
        self.dataset_id = dataset_id
        self.project = project
        self.location = location

    def get_dataset(self) -> aiplatform.ImageDataset:
        aiplatform.init(project=self.project, location=self.location)
        ds = aiplatform.ImageDataset(self.dataset_id)
        return ds

    def export_data_files(self, output_dir: str) -> Sequence[str]:
        ds = self.get_dataset()
        data_files = ds.export_data(output_dir=output_dir)
        return data_files



import json

from captur_ml_sdk.dtypes.exceptions import GoogleCloudPubSubTopicDoesNotExistError
from google.cloud import pubsub_v1
from google.api_core import exceptions as google_exceptions

def publish(
    payload: dict,
    topic_name: str,
    project_name: str = "capturpwa",
    publisher: pubsub_v1.PublisherClient = None,
):
    """Publishes a message to a Pub/Sub topic.

    Args:
        payload (dict): Jsonlike message to be published.
        topic_name (str): Name of the Pub/Sub topic to publish to.
        project_name (str, optional): The GCP project name. Defaults to "capturpwa".
        publisher (pubsub_v1.PublisherClient, optional): The Publisher Client. Defaults to None.

    Raises:
        GoogleCloudPubSubTopicDoesNotExistError: Raised if the topic is not found.
    """
    if publisher is None:
        publisher = pubsub_v1.PublisherClient()

    data = json.dumps(payload).encode("utf-8")

    topic_path = publisher.topic_path(
        project_name, topic_name
    )

    future = publisher.publish(
        topic_path, data
    )

    try:
        future.result()
    except google_exceptions.NotFound:
        raise GoogleCloudPubSubTopicDoesNotExistError()


def test_topic_permissions(
    project_id,
    topic_id,
    client: pubsub_v1.PublisherClient = None,
):
    if client is None:
        client = pubsub_v1.PublisherClient()
    topic_path = client.topic_path(project_id, topic_id)

    permissions_to_check = ["pubsub.topics.publish", "pubsub.topics.update"]

    allowed_permissions = client.test_iam_permissions(
        request={"resource": topic_path, "permissions": permissions_to_check}
    )

    print(
        "Allowed permissions for topic {}: {}".format(topic_path, allowed_permissions)
    )


# class GC_Pubsub(object):
#     def __init__(self):
#         self._client = pubsub_v1.PublisherClient()

#     def publish(
#         self,
#         payload: Dict,
#         project_name: str,
#         topic_name: str
#     ):

#         try:
#             publish(
#                 payload=payload,
#                 project_name=project_name,
#                 topic_name=topic_name,
#                 publisher=self._client
#             )

#     def test_topic_permissions(
#         self,
#         project_id: str,
#         topic_id: str
#     ):
#         test_topic_permissions(
#             project_id=project_id,
#             topic_id=topic_id,
#             client=self._client
#         )



from setuptools import setup, find_packages

long_desc = open("README.md").read()
required = [
    'pydantic',
    'google-cloud-storage',
    'google-cloud-aiplatform',
    'google-cloud-pubsub',
    'google-auth',
    'sentry_sdk',
]

setup(
    name='captur-ml',
    version='0.2.24',
    description='The internal Captur Machine Learning SDK',
    long_description=long_desc,
    long_description_content_type="text/markdown",
    url='https://github.com/Captur/captur-ml-sdk',
    author=[
        'Jack Barnett Leveson',
        'Jonny Jackson'
    ],
    author_email='jack@capturphotos.com',
    license='Creative Commons Attribution-NonCommercial-NoDerivatives 4.0 International',
    packages=find_packages(exclude=("test",)),
    install_requires=required,
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    python_requires=">=3.7"
)



import json
from typing import Any, Dict, List, Union, Literal, Optional

from google.cloud import storage
from google.auth import exceptions as ga_exceptions
from google.api_core import exceptions as google_exceptions

from captur_ml_sdk.dtypes.exceptions import (
    GoogleCloudStoragePermissionError,
    GoogleCloudStorageResourceNotFoundError,
    GoogleCloudStorageBucketNotFoundError,
)


def _get_storage_client() -> storage.Client:
    """Wrapper function that returns an instance of the Google Cloud Storage client with added Exception handling.

    Raises:
        GoogleCloudStoragePermissionError: If the user does not have permission to access the Google Cloud Storage bucket, for example if there is
            no credentials file or the credentials file is invalid.

    Returns:
        storage.Client: A Google Cloud Storage client instance.
    """
    try:
        storage_client = storage.Client()
    except ga_exceptions.DefaultCredentialsError:
        raise GoogleCloudStoragePermissionError("Failed to create Google Cloud Storage client. Google Cloud Storage permissions are not configured.")

    return storage_client


def write_file_to_gcs(file_contents: Union[str, bytes], bucket_name: str, destination_filename: str, storage_client: Optional[storage.Client] = None):
    """Writes a file of either string or bytes content type to Google Cloud Storage.

    Args:
        file_contents (Union[str, bytes]): The contents of the file to write to Google Cloud Storage.
        bucket_name (str): The name of the Google Cloud Storage bucket to write the file to.
        destination_filename (str): The name of the file to write to Google Cloud Storage.
        storage_client (Optional[storage.Client], optional): A Google Cloud Storage client instance. If not provided, a new instance will be created. Defaults to None.

    Raises:
        GoogleCloudStorageResourceNotFoundError: If the Google Cloud Storage bucket does not exist.
    """
    if storage_client is None:
        storage_client = _get_storage_client()

    bucket = storage_client.bucket(bucket_name)
    blob = bucket.blob(destination_filename)

    try:
        blob.upload_from_string(file_contents)
    except google_exceptions.NotFound:
        raise GoogleCloudStorageResourceNotFoundError(f"Failed to write file to Google Cloud Storage. Bucket {bucket_name} not found.")


def write_json_to_gcs(json_dict: Dict[Any, Any], bucket_name: str, destination_filename: str, storage_client: Optional[storage.Client] = None):
    """Writes a JSON file to Google Cloud Storage.

    Args:
        json_dict (Dict[Any, Any]): The dictionary to export as JSON to Google Cloud Storage.
        bucket_name (str): The name of the Google Cloud Storage bucket to write the file to.
        destination_filename (str): The name of the file to write to Google Cloud Storage. If it does not end with the .json extension, it will be added.
        storage_client (Optional[storage.Client], optional): A Google Cloud Storage client instance. If not provided, a new instance will be created. Defaults to None.

    Raises:
        ValueError: If the dictionary cannot be serialized to JSON.
    """
    if not destination_filename.endswith(".json"):
        destination_filename = f"{destination_filename}.json"

    try:
        json_string = json.dumps(json_dict)
        json.loads(json_string)
    except json.JSONDecodeError:
        raise ValueError("Failed to write JSON to Google Cloud Storage. JSON data is not valid.")

    write_file_to_gcs(json_string, bucket_name=bucket_name, destination_filename=destination_filename, storage_client=storage_client)


def write_jsonl_to_gcs(json_lines: List[Dict], bucket_name: str, destination_filename: str, storage_client: Optional[storage.Client] = None):
    """Writes a JSON file to Google Cloud Storage.

    Args:
        json_lines (List[Dict]): List of dictionaries to export as JSON lines to Google Cloud Storage.
        bucket_name (str): The name of the Google Cloud Storage bucket to write the file to.
        destination_filename (str): The name of the file to write to Google Cloud Storage. If it does not end with the .json extension, it will be added.
        storage_client (Optional[storage.Client], optional): A Google Cloud Storage client instance. If not provided, a new instance will be created. Defaults to None.

    Raises:
        ValueError: If the dictionary cannot be serialized to JSON.
    """
    if not destination_filename.endswith(".jsonl"):
        destination_filename = f"{destination_filename}.jsonl"

    try:
        json_lines_string = ""
        for line in json_lines:
            json_lines_string += json.dumps(line) + "\n"
    except json.JSONDecodeError:
        raise ValueError("Failed to write JSON to Google Cloud Storage. JSON data is not valid.")

    write_file_to_gcs(json_lines_string, bucket_name=bucket_name, destination_filename=destination_filename, storage_client=storage_client)


def read_file_from_gcs(bucket_name: str, origin_filename: str, read_mode: Literal["r", "b"] = "r", storage_client: Optional[storage.Client] = None) -> Union[bytes, str]:
    """Reads a file from Google Cloud Storage.

    Args:
        bucket_name (str): The name of the Google Cloud Storage bucket to read the file from.
        origin_filename (str): The name of the file to read from Google Cloud Storage.
        read_mode (Literal["r", "b"], optional): The mode to read the file in. "r" for text, "b" for bytes. Defaults to "r".
        storage_client (Optional[storage.Client], optional): A Google Cloud Storage client instance. If not provided, a new instance will be created. Defaults to None.

    Raises:
        GoogleCloudStorageResourceNotFoundError: If the Google Cloud Storage bucket or file does not exist.
        ValueError: If the read_mode is not "b" but the file cannot be decoded as UTF-8.

    Returns:
        Union[bytes, str]: The raw contents of the file.
    """
    if storage_client is None:
        storage_client = _get_storage_client()

    bucket = storage_client.bucket(bucket_name)
    blob = bucket.blob(origin_filename)

    try:
        contents = blob.download_as_bytes()
    except google_exceptions.NotFound:
        raise GoogleCloudStorageResourceNotFoundError(f"Failed to read file from Google Cloud Storage. File {bucket_name}/{origin_filename} not found.")

    if read_mode != "b" and type(contents) is bytes:
        try:
            contents = contents.decode("utf-8")
        except UnicodeDecodeError:
            raise ValueError(f"Failed to decode file-bytes from Google Cloud Storage. File {bucket_name}/{origin_filename} is not a valid UTF-8 string. Please use read_mode='b' to read bytes.")

    return contents


def delete_file_from_gcs(bucket_name: str, destination_filename: str, storage_client: Optional[storage.Client] = None):
    """Deletes a file from Google Cloud Storage.

    Args:
        bucket_name (str): The name of the Google Cloud Storage bucket to delete the file from.
        destination_filename (str): The name of the file to delete from Google Cloud Storage.
        storage_client (Optional[storage.Client], optional): A Google Cloud Storage client instance. If not provided, a new instance will be created. Defaults to None.

    Raises:
        GoogleCloudStorageResourceNotFoundError: If the Google Cloud Storage bucket or file does not exist.
    """
    if storage_client is None:
        storage_client = _get_storage_client()

    bucket = storage_client.bucket(bucket_name)
    blob = bucket.blob(destination_filename)

    try:
        blob.delete()
    except google_exceptions.NotFound:
        raise GoogleCloudStorageResourceNotFoundError(f"Failed to delete file from Google Cloud Storage. File {bucket_name}/{destination_filename} not found.")


def check_bucket_exists(bucket_name: str, storage_client: Optional[storage.Client] = None) -> bool:
    """Checks if a Google Cloud Storage bucket exists.

    Args:
        bucket_name (str): The name of the Google Cloud Storage bucket to check.
        storage_client (Optional[storage.Client], optional): A Google Cloud Storage client instance. If not provided, a new instance will be created. Defaults to None.

    Returns:
        bool: True if the bucket exists, False otherwise.
    """
    if storage_client is None:
        storage_client = _get_storage_client()

    bucket = storage_client.bucket(bucket_name)
    return bucket.exists()


def check_file_exists(bucket_name: str, file_name: str, storage_client: Optional[storage.Client] = None) -> bool:
    """Checks if a file exists in Google Cloud Storage.

    Args:
        gcs_filename (str): The name of the file to check.
        storage_client (Optional[storage.Client], optional): A Google Cloud Storage client instance. If not provided, a new instance will be created. Defaults to None.

    Returns:
        bool: True if the file exists, False otherwise.
    """
    if storage_client is None:
        storage_client = _get_storage_client()

    if not check_bucket_exists(bucket_name):
        raise GoogleCloudStorageBucketNotFoundError(f"{bucket_name} does not exist.")

    bucket = storage_client.bucket(bucket_name)
    blob = bucket.blob(file_name)

    return blob.exists()


class GCS_Storage(object):
    """A class for interacting with Google Cloud Storage.
    """
    def __init__(self):
        """Initialises the Google Cloud Storage client with additional Exception handling.

        Raises:
            GoogleCloudStoragePermissionError: If the user does not have permission to access the Google Cloud Storage bucket, for example if there is
                no credentials file or the credentials file is invalid.

        Returns:
            storage.Client: A Google Cloud Storage client instance.
        """
        self._client = _get_storage_client()

    def write(self, file_contents: Union[str, bytes], bucket_name: str, destination_filename: str):
        """Writes a file of either string or bytes content type to Google Cloud Storage.

        Args:
            file_contents (Union[str, bytes]): The contents of the file to write to Google Cloud Storage.
            bucket_name (str): The name of the Google Cloud Storage bucket to write the file to.
            destination_filename (str): The name of the file to write to Google Cloud Storage.

        Raises:
            GoogleCloudStorageResourceNotFoundError: If the Google Cloud Storage bucket does not exist.
        """
        write_file_to_gcs(file_contents=file_contents, bucket_name=bucket_name, destination_filename=destination_filename, storage_client=self._client)

    def write_json(self, json_dict: Dict[Any, Any], bucket_name: str, destination_filename: str):
        """Writes a JSON file to Google Cloud Storage.

        Args:
            json_dict (Dict[Any, Any]): The dictionary to export as JSON to Google Cloud Storage.
            bucket_name (str): The name of the Google Cloud Storage bucket to write the file to.
            destination_filename (str): The name of the file to write to Google Cloud Storage. If it does not end with the .json extension, it will be added.

        Raises:
            ValueError: If the dictionary cannot be serialized to JSON.
        """
        write_json_to_gcs(json_dict=json_dict, bucket_name=bucket_name, destination_filename=destination_filename, storage_client=self._client)

    def read(self, bucket_name: str, origin_filename: str, read_mode: Literal["r", "b"] = "r") -> Union[bytes, str]:
        """Reads a file from Google Cloud Storage.

        Args:
            bucket_name (str): The name of the Google Cloud Storage bucket to read the file from.
            origin_filename (str): The name of the file to read from Google Cloud Storage.
            read_mode (Literal["r", "b"], optional): The mode to read the file in. "r" for text, "b" for bytes. Defaults to "r".

        Raises:
            GoogleCloudStorageResourceNotFoundError: If the Google Cloud Storage bucket or file does not exist.
            ValueError: If the read_mode is not "b" but the file cannot be decoded as UTF-8.

        Returns:
            Union[bytes, str]: The raw contents of the file.
        """
        return read_file_from_gcs(bucket_name=bucket_name, origin_filename=origin_filename, read_mode=read_mode, storage_client=self._client)

    def delete(self, bucket_name: str, destination_filename: str):
        """Deletes a file from Google Cloud Storage.

        Args:
            bucket_name (str): The name of the Google Cloud Storage bucket to delete the file from.
            destination_filename (str): The name of the file to delete from Google Cloud Storage.

        Raises:
            GoogleCloudStorageResourceNotFoundError: If the Google Cloud Storage bucket or file does not exist.
        """
        delete_file_from_gcs(bucket_name=bucket_name, destination_filename=destination_filename, storage_client=self._client)



__version__ = "0.2.24"


# from ._services.aiplatform.live_predict import get_image_classification_prediction_from_deployed_automl

# __all__ = ["get_image_classification_prediction_from_deployed_automl"]


import pydantic
from typing import List

from captur_ml_sdk.dtypes.generics import Image, Prediction
from google.cloud import aiplatform_v1


class ImageClassificationResult(pydantic.BaseModel):
    id: str = pydantic.Field(
        ...,
        description="The image used for the prediction.",
    )
    predictions: List[Prediction] = pydantic.Field(
        ...,
        description="The list of classification results for the image.",
    )

    @classmethod
    def from_automl_response(cls, image: Image, automl_response_object: aiplatform_v1.types.PredictResponse):
        # models_predictions will have one item for each model that has had predictions requested from it.
        #       In many situations it will therefore only have a single item.
        models_predictions = [dict(prediction) for prediction in automl_response_object.predictions]

        predictions = []
        for prediction in models_predictions:
            ids = prediction["ids"]
            class_names = prediction["displayNames"]
            confidences = prediction["confidences"]
            # ensure predictions are in descending order of confidence
            arr = [(conf, cls, id) for conf, cls, id in zip(confidences, class_names, ids)]
            arr.sort(reverse=True, key=lambda x: x[0])
            confidences = [x[0] for x in arr]
            class_names = [x[1] for x in arr]
            ids = [x[2] for x in arr]
            for id, name, confidence in zip(ids, class_names, confidences):
                pred = Prediction(id=id, displayName=name, confidence=confidence)
                predictions.append(pred)

        return cls(id=image.id, predictions=predictions)
